import unittest
from mergesvvcf.locations import *
from mergesvvcf.variantdict import *

class TestLocations(unittest.TestCase):

    def setUp(self):
        self.l1 = location('X',31,'+')
        self.l2 = location('Y',131,-1)
        self.l3 = location('Y',121,+1)
        self.l4 = location('Y',161,'-')
        self.ld = locationdict(window=40)
        self.ld[self.l1] = 'foo'
        self.ld[self.l2] = str(self.l2)
        self.ld[self.l3] = 'baz'

    def test_comparisons(self):
        self.assertTrue( self.l3 < self.l2 )
        self.assertTrue( self.l4 > self.l2 )
        self.assertFalse( self.l3.overlap(self.l4,strandTest=True,window=40) )
        self.assertTrue( self.l2.overlap(self.l4,strandTest=True,window=40) )

    def test_locationdict(self):
        self.assertTrue( self.l4 in self.ld )  # should overlap with l2
        # self.assertTrue( self.ld.__str__() == "{(Y,121): 'baz', (Y,131,-L): '(Y,131,-L)', (X,31): 'foo'}") # str method of dict is not deterministic!
        self.assertTrue( self.ld[self.l4].__str__() == "(Y,131,-L)")
        self.assertTrue( self.l3 in self.ld )
        self.assertTrue( self.l2 in self.ld )
        self.assertTrue( self.l1 in self.ld )

class TestVariantMap(unittest.TestCase):

    def setUp(self):
        self.l1 = location('X',31,'+')
        self.l2 = location('Y',131,-1)
        self.l3 = location('Y',121,+1)
        self.l4 = location('Y',161,'-')
        self.vmap = variantmap(0,40)
        self.vmap[(self.l1,('A','G'))] = 'sga'
        self.vmap[(self.l2,('ACGTA','A'))] = 'sanger'
        self.vmap[(self.l1,('A','G'))] = 'dkfz'
        self.vmap[(self.l2,self.l3)] = 'someone_else'
        self.vmap[(self.l4,self.l3)] = 'different'
        self.vmap[(self.l2,None)] = 'loosend'

    def test_vardict(self):
        self.assertTrue( (self.l1,('A','G')) in self.vmap )
        self.assertTrue( (self.l1,('C','G')) not in self.vmap )
        self.assertTrue( (self.l2,self.l3) in self.vmap )

    def test_iteration(self):
        nin = 0
        for var in self.vmap:
            nin += 1
        self.assertTrue( nin == 4 )

if __name__ == '__main__':
    unittest.main()



"""
Definitions for a dictionary of variants, and operations on them
"""
from mergesvvcf.locations import locationdict, location
import pysam
import mergesvvcf.vcftobreakpoints as svvcf

def __checkvalidpairlocs__(t):
    """
    Returns true if t is a tuple of 2 locations, false otherwise
    """
    if type(t) is not tuple:
        return False
    if len(t) != 2:
        return False
    if type(t[0]) is location and type(t[1]) is location:
        return True
    return False

class locationpairdict(object):
    def __init__(self, window):
        self.__window = window
        self.__lpdict = locationdict(self.__window)

    def __contains__(self, lpair):
        if not __checkvalidpairlocs__(lpair):
            raise KeyError("Required: tuple of locations")
        locn1 = lpair[0]; locn2 = lpair[1]
        if not locn1 in self.__lpdict:
            return False
        return locn2 in self.__lpdict[locn1]

    def __getitem__(self, lpair):
        if type(lpair) is location:
            return self.__lpdict[lpair]

        if not __checkvalidpairlocs__(lpair):
            raise KeyError("Required: tuple of locations")
        if not self.__contains__(lpair):
            raise KeyError("Not in location pair dict")

        locn1 = lpair[0]; locn2 = lpair[1]
        return self.__lpdict[locn1].get(locn2)

    def __setitem__(self, lpair, entry):
        if not __checkvalidpairlocs__(lpair):
            raise KeyError("Required: tuple of locations")
        locn1 = lpair[0]
        locn2 = lpair[1]
        if not locn1 in self.__lpdict:
            self.__lpdict[locn1] = locationdict(self.__window)
        if not locn2 in self.__lpdict[locn1]:
            self.__lpdict[locn1][locn2] = []
        self.__lpdict[locn1][locn2].append(entry)

    def keys(self):
        return list(self.__lpdict.keys())

    def __iter__(self):
        return self.__lpdict.__iter__()

class variantmap(object):
    def __init__(self, awindow, svwindow):
        self.__awindow = awindow
        self.__svwindow = svwindow

        self.__alleledict = locationdict(awindow)     # map locn -> allele (ref/alt)
        self.__svdict = locationpairdict(svwindow)    # map locn -> locn (for SVs - paired breakpoints)
        self.__locn1pos = locationpairdict(svwindow)  # map locn -> all the locations found which map to this one
        self.__locn2pos = locationpairdict(svwindow)  # map locn -> all the locations found which map to this one
        self.__records = locationpairdict(svwindow)   # all the records involving this location

    def __medianpos__(self, locn1, locn2):
        def median(l):
            mid = len(l) // 2
            ls = sorted(l)
            if len(ls) % 2 == 0:
                return (ls[mid-1] + ls[mid]) // 2
            else:
                return ls[mid]

        if not (locn1, locn2) in self.__svdict:
            return None, None
        posn1s = self.__locn1pos[(locn1, locn2)]
        posn2s = self.__locn2pos[(locn1, locn2)]
        return median(posn1s), median(posn2s)

    def __svpresent__(self, locn1, locn2):
        return (locn1, locn2) in self.__svdict

    def __allelepresent__(self, locn, allele):
        if not locn in self.__alleledict:
            return False
        return allele in self.__alleledict[locn]

    def __addsvcaller__(self, locn1, locn2, caller):
        self.__svdict[(locn1, locn2)] = caller

    def __addallelecaller__(self, locn, allele, caller):
        if not locn in self.__alleledict:
            self.__alleledict[locn] = {}
        if not allele in self.__alleledict[locn]:
            self.__alleledict[locn][allele] = []
        if not caller in self.__alleledict[locn][allele]:
            self.__alleledict[locn][allele].append(caller)

    def __contains__(self, vartuple):
        assert type(vartuple) is tuple or type(vartuple) is list
        assert len(vartuple) > 1
        locn = vartuple[0]
        other = vartuple[1]
        if type(other) is location:
            return self.__svpresent__(locn, other)
        else:
            return self.__allelepresent__(locn, other)

    def __setitem__(self, vartuple, caller, record=None):
        """Add item to the appropriate dictionary, and update locations
        dictionary as well."""
        assert type(vartuple) is tuple or type(vartuple) is list
        assert len(vartuple) > 1
        locn = vartuple[0]
        other = vartuple[1]
        if other is None:
            other = location(None,0)
        if type(other) is location:
            self.__addsvcaller__(locn, other, caller)
            self.__locn1pos[(locn,other)] = locn.__pos__
            self.__locn2pos[(locn,other)] = other.__pos__
            self.__records[(locn,other)] = (caller, record)
        else:
            self.__addallelecaller__(locn, other, caller)

    def __getitem__(self, vartuple):
        assert type(vartuple) is tuple or type(vartuple) is list
        assert len(vartuple) > 1
        if not self.__contains__(vartuple):
            raise KeyError(str(vartuple))
        else:
            locn = vartuple[0]
            other = vartuple[1]
            if type(other) is location:
                return self.__svdict[locn][other]
            else:
                return self.__alleledict[locn][other]

    def __str__(self):
        output=""
        for loc in self.__alleledict:
            chrom = loc.__chrom__
            pos = loc.__pos__
            for item in self.__alleledict[loc]:
                ref, alt = item
                output+="\t".join([chrom, str(pos), '.', ref, alt])
                output+="\tCallers="+",".join(self.__alleledict[loc][item])+"\n"

        for loc1 in self.__svdict:
            chrom1 = loc1.__chrom__
            for loc2 in self.__svdict[loc1]:
                chrom2 = loc2.__chrom__
                pos1, pos2 = self.__medianpos__(loc1, loc2)
                output+="\t".join([chrom1, str(pos1), '.', 'n', 'n['+chrom2+":"+str(pos2)+'['])
                output+="\tCallers="+",".join(self.__svdict[loc1][loc2])+"\n"
        return output

    def __repr__(self):
        return self.__str__()

    # forceSV is here to allow forcing a call that looks like a huge indel to be treated as an SV
    def addrecord(self, record, caller="NA", forceSV=False):
        assert type(record) is pysam.VariantRecord

        if forceSV or len(list(record.alts)) > 0:
            bkptPairs = svvcf.breakpointsFromRecord(record)
            for pair in bkptPairs:
                self.__setitem__(pair, caller, record)
        else:
            for alt in record.alts:
                if alt is None:
                    continue
                loc = location(record.chr, int(record.pos))
                allele = (record.ref, str(alt))
                self.__setitem__((loc, allele), caller)

    def __iter__(self):
        def generatorIterator():
            for loc in self.__alleledict:
                for allele in self.__alleledict[loc]:
                    yield loc, allele, self.__alleledict[loc][allele]
            for loc1 in self.__svdict:
                for loc2 in self.__svdict[loc1]:
                    pos1, pos2 = self.__medianpos__(loc1, loc2)
                    yield loc1, loc2, self.__svdict[loc1][loc2], pos1, pos2, self.__records[(loc1,loc2)]
            raise StopIteration()

        return generatorIterator()



import mergesvvcf.mergedfile as mergedfile
import argparse
import os
import sys

def main():
    """Merge SV VCF files, output to stdout or file"""
    defsvwindow = 100

    parser = argparse.ArgumentParser(description='Merge calls in SV VCF files')
    parser.add_argument('input_files', nargs='+', help='Input VCF files')
    parser.add_argument('-o', '--output', type=argparse.FileType('w'), default=sys.stdout, help="Specify output file (default:stdout)")
    parser.add_argument('-v', '--verbose', action='store_true', help="Specify verbose output")
    parser.add_argument('-l', '--labels', type=str, help='Comma-separated labels for each input VCF file (default:basenames)')
    parser.add_argument('-n', '--ncallers', action='store_true', help='Annotate variant with number of callers')
    parser.add_argument('-m', '--mincallers', type=int, default=0, help='Minimum # of callers for variant to pass')
    parser.add_argument('-s', '--sv', action='store_true', help='Force interpretation as SV (default:false)')
    parser.add_argument('-f', '--filtered', action='store_true', help='Include records that have failed one or more filters (default:false)')
    parser.add_argument('-w', '--svwindow', default=defsvwindow, type=int,
                         help='Window for comparing breakpoint positions for SVs (default:'+str(defsvwindow)+')')
    parser.add_argument('-d', '--debug', action=('store_true'),
                         help='Include original reads and raise errors (default:false)')

    args = parser.parse_args()
    input_files = args.input_files
    if args.labels is None:
        labels = [os.path.splitext(os.path.basename(f))[0] for f in input_files]
    else:
        labels = [label.strip() for label in args.labels.split(',')]

    mergedfile.merge(input_files, labels, args.sv, args.output,
                     slop=args.svwindow, verbose=args.verbose,
                     output_ncallers=args.ncallers,
                     min_num_callers=args.mincallers,
                     filterByChromosome=True, noFilter=args.filtered,
                     debug=args.debug)



"""
Classes for signed genomic locations - eg, endpoints of half-intervals
"""

class location(object):
    __strands = {+1:True, -1:False, "+":True, "-":False, True:True, False:False}
    __strandstrs = {True:"+", False:"-"}
    __extendstrs = {True:"R", False:"L"}
    def __init__(self, chrom, pos, strand=True, extendsRight=False):
        assert type(pos) is int
        assert strand in self.__strands
        assert type(extendsRight) is bool
        self.__chrom__ = str(chrom)
        self.__pos__ = pos
        self.__strand__ = self.__strands[strand]
        self.__right__ = extendsRight

    def __hash__(self):
        return hash(self.__chrom__) ^ hash(self.__pos__) ^ hash(self.__strand__) ^ hash(self.__right__)

    def __lt__(self, other):
        """(Less than) Compare location of two positions, regardless of strand or direction."""
        def _strsAsIntsIfPossible(str1, str2):
            try:
                i1 = int(str1)
                i2 = int(str2)
            except ValueError:
                return str1, str2
            return i1, i2

        if other is None:
            return -1
        if self.__chrom__ != other.__chrom__:
            ic1, ic2 = _strsAsIntsIfPossible(self.__chrom__, other.__chrom__)
            return ic1 < ic2
        return self.__pos__ < other.__pos__

    def __eq__(self, other):
        """(Equal) Compare location of two positions, regardless of strand or direction."""
        def _strsAsIntsIfPossible(str1, str2):
            try:
                i1 = int(str1)
                i2 = int(str2)
            except ValueError:
                return str1, str2
            return i1, i2

        if other is None:
            return -1
        if self.__chrom__ != other.__chrom__:
            ic1, ic2 = _strsAsIntsIfPossible(self.__chrom__, other.__chrom__)
            return ic1 == ic2
        return self.__pos__ == other.__pos__

    def overlap(self, other, strandTest=False, window=0):
        """Compare location of two endpoints with the given window, optionally
        testing for strand"""
        if other is None:
            return False
        if self.__chrom__ != other.__chrom__:
            return False
        if strandTest and (self.__strand__ != other.__strand__):
            return False
        return abs(self.__pos__ - other.__pos__) <= window

    def __add__(self, offset):
        """Add offset to position"""
        return location(self.__chrom__, self.__pos__+offset, self.__strand__, self.__right__)

    def __str__(self):
        """Convert to string"""
        return '(%s,%d,%s%s)' % (self.__chrom__, self.__pos__, self.__strandstrs[self.__strand__], self.__extendstrs[self.__right__])

    def __repr__(self):
        """Representation for printing"""
        extrastr=""
        if (not self.__strand__) or (self.__right__):
            extrastr = ",%s%s" % (self.__strandstrs[self.__strand__], self.__extendstrs[self.__right__])
        return '(%s,%d%s)' % (self.__chrom__, self.__pos__, extrastr)

    def rc(self):
        """Reverse strand of current strand"""
        return location(self.__chrom__, self.__pos__, not self.__strand__, self.__right__)

    def isRC(self):
        """Is on reverse strand?"""
        return self.__strand__ == False

    def switchExtent(self):
        return location(self.__chrom__, self.__pos__, self.__strand__, not self.__right__)

    def asTuple(self):
        return self.__chrom__, self.__pos__, self.__strand__, self.__right__

    def withPos(self, pos):
        """Replace position of location"""
        return location(self.__chrom__, pos, self.__strand__, self.__right__)

    @property
    def pos(self):
        return self.__pos__

    @property
    def chrom(self):
        return self.__chrom__

class locationdict(dict):
    def __init__(self, window, *args, **kwargs):
        self.__window = window
        self.__search = [0] + [item for pm in zip(list(range(1,window+1)),
                                                  list(range(-1,-window-1,-1)))
                                                for item in pm]
        super(locationdict,self).__init__(*args, **kwargs)

    def keys(self):
        return list(super(locationdict, self).keys())

    def values(self):
        return [self[key] for key in self]

    def itervalues(self):
        return (self[key] for key in self)

    def __find__(self, locn):
        if not type(locn) is location:
            raise ValueError("Not Location: "+locn.__str__())
        present = False
        foundoff = None
        for off in self.__search:
            if super(locationdict, self).__contains__(locn+off):
                present = True
                foundoff = off
                break
        return present, foundoff

    def __contains__(self, locn):
        if not type(locn) is location:
            raise ValueError("Not Location: "+locn.__str__())
        present, foundloc = self.__find__(locn)
        return present

    def __getitem__(self, locn):
        if not type(locn) is location:
            raise ValueError("Not Location: "+locn.__str__())
        present, foundoff = self.__find__(locn)
        if not present:
            raise KeyError(locn.__str__())
        else:
            return super(locationdict, self).__getitem__(locn + foundoff)



#!/usr/bin/env python
from __future__ import print_function
import argparse
import sys
import re
import pysam
import mergesvvcf.locations as loc

__symbolicRE__ = None
__bpRE__ = None
__looseendRE__ = None

def setupREs():
    global __symbolicRE__
    global __bpRE__
    global __looseendRE__
    if __symbolicRE__ is None or __bpRE__ is None or __looseendRE__ is None:
        __symbolicRE__ = re.compile(r'.*<([A-Z:]+)>.*')
        __bpRE__ = re.compile(r'([ACGTNactgn\.]*)([\[\]])([a-zA-Z0-9_\.]+:\d+)([\[\]])([ACGTNacgtn\.]*)')
        __looseendRE__ = re.compile(r'[lL][oO][oO][sS][eE][eE][nN][dD]')

def stdchrom(chrom):
    if chrom[0]=='c':
        return chrom[3:]
    else:
        return chrom

def orderBreakpoints(loc1, loc2):
    # already ordered?  Just return them
    if loc1 < loc2:
        return loc1, loc2

    # Otherwise, swap and either flip strands (if different) or extents
    first, second = loc2, loc1
    if first.isRC():
        first, second = first.rc(), second.rc()

    return (first, second)

def otherPosnSymbolic(info):

    def getIfPresent(name, infodict):
        res = None
        if name in infodict:
            res = infodict[name]
            if type(res) is list or type(res) is tuple:
                res = res[0]
        return res

    fields = ['CHR2','END','CT','SVTYPE','SVLEN']
    results = []
    for field in fields:
        results.append( getIfPresent(field, info) )

    # handle chr2 specially
    if results[0] is not None:
        results[0] = stdchrom(results[0])

    # svclass overrides svtype
    svclass = getIfPresent('SVCLASS', info)
    if svclass is not None:
        results[3] = svclass

    transdict = {"inversion":"INV", "deletion":"DEL",
            "tandem_dup":"DUP:TANDEM", "duplication":"DUP",
            "long_range":"TRA", "inter_chr":"TRA"}

    if results[3] in transdict:
        results[3] = transdict[results[3]]

    return tuple(results)

def ctAndLocFromBkpt(ref, pre, delim1, pair, delim2, post):
    # extract strand/orientation, position, and (possibly) inserted string
    # length from record eg [9:1678896[N.
    # result is result from re.match with the explicit BP regexp

    chpos=pair.split(':')
    chr2 = stdchrom(chpos[0])
    pos2 = int(chpos[1])
    assert delim1 == delim2  # '['..'[' or ']'...']'
    joinedAfter = True
    extendRight = True
    connectSeq = ""

    if len(pre) > 0:
        connectSeq = pre
        joinedAfter = True
        assert len(post) == 0
    elif len(post) > 0:
        connectSeq = post
        joinedAfter = False

    if delim1 == "]":
        extendRight = False
    else:
        extendRight = True

    indellen = len(connectSeq) - len(ref)

    if joinedAfter:
        if extendRight:
            ct = '3to5'
        else:
            ct = '3to3'
    else:
        if extendRight:
            ct = '5to5'
        else:
            ct = '5to3'

    return ct, chr2, pos2, indellen

def translocation(chr1, pos1, chr2, pos2, ct):
    secondStrand = "+"
    firstExtendRight = False
    secondExtendRight = True
    if ct is None or ct == '3to5':
        pass
    elif ct == '5to3':
        firstExtendRight = True; secondExtendRight = False
    elif ct == '5to5':
        firstExtendRight = True; secondExtendRight = True; secondStrand = '-'
    elif ct == '3to3':
        firstExtendRight =False; secondExtendRight =False; secondStrand = '-'

    return ( loc.location(chr1, pos1, "+",          firstExtendRight),
             loc.location(chr2, pos2, secondStrand, secondExtendRight) )

def breakpointsFromRecord(record):
    """Returns a list of pair(s) of breakpoints corresponding to the record."""
    global __symbolicRE__
    global __bpRE__
    global __looseendRE__

    chr1, pos1 = stdchrom(record.chrom), int(record.pos)
    first = loc.location(chr1, pos1)

    if record.alts is None:
        return [(first, loc.location(None,0,"+",True))]

    setupREs()
    ref = str(record.ref)
    bkptPairs = []

    for alt in record.alts:
        if alt is None:
            altstr = "."
        else:
            altstr = str(alt)

        # get all available information from the record
        chr2, pos2, ct, svtype, svlen = otherPosnSymbolic(record.info)

        # defaults
        if chr2 is None:
            chr2 = chr1
        if pos2 is None and "<" in altstr:
            pos2 = record.stop

        # look for symbolic SVTYPE information in the alt field (eg, <DEL>)
        resultSym = re.match(__symbolicRE__, altstr)
        if resultSym:
            svtype = resultSym.group(1)

        # explicit BP - look for chr2 and CT information from the alt field
        # (eg, N[chr2:123123[)
        resultBP = re.match(__bpRE__, altstr)
        if resultBP:
            ct, chr2, pos2, indellen = ctAndLocFromBkpt(str(record.ref), resultBP.group(1),
                    resultBP.group(2), resultBP.group(3), resultBP.group(4),
                    resultBP.group(5))
            if svlen is None:
                svlen = indellen

        # looseend; no paired BP
        resultLE = __looseendRE__.search(str(record.filter))
        if altstr == ref+"." or resultLE:
            chr2 = None; pos2 = 0
            if ct is None:
                ct = '5to3'
        if altstr == "."+ref:
            chr2 = None; pos2 = 0
            if ct is None:
                ct = '3to5'

        # if nothing else, treat as an indel
        if (not svtype or svtype == "BND") and not resultBP and not resultLE and not resultSym:

            reflen = len(ref)
            if pos2 is None:
                pos2 = pos1 + reflen
            if svlen is None:
                svlen = len(altstr) - reflen
            if svtype is None:
                if svlen > 0:
                    svtype = 'INS'
                else:
                    svtype = 'DEL'

        if svtype == "INV":
            assert chr1 == chr2
            # Only one version of INV reporting
            bkptPairs.append(translocation(chr1, pos1, chr1, pos2, ct))
            # bkptPairs.append( translocation(chr1, pos1,   chr1, pos2-1, '3to3') )
            # bkptPairs.append( translocation(chr1, pos1+1, chr1, pos2,   '5to5') )
        elif svtype in ["DUP:TANDEM","DUP"]:
            if chr2 is None:
                chr2 = chr1
            if ct is None or ct=="5to3" or ct=="3to5":
                if pos1 > pos2:
                    pos1, pos2 = pos2, pos1
                bkptPairs.append( translocation(chr1, pos1, chr2, pos2, "5to3") )
            else:
                bkptPairs.append( translocation(chr1, pos1, chr2, pos2, ct) )
        elif svtype in ["INS","INS:ME:L1"]:
            assert chr1 == chr2
            bkptPairs.append((first, loc.location(chr1, pos1+1, "+", True)))
        elif svtype in ["DEL:ME:ALU","DEL"]:
            assert chr1 == chr2
            bkptPairs.append( translocation(chr1, min(pos1,pos2), chr1, max(pos1,pos2), '3to5') )
        else:
            # translocation is the default
            if ct is None:
                ct = "3to5"
            # if (svtype is not None) and (not svtype in ["TRA","BND"]):
            #     print("Got unknown record of type", svtype, altstr, str(record), file=sys.stderr)
            #     print("Hoping for best",file=sys.stderr)
            bkptPairs.append( translocation(chr1, pos1, chr2, pos2, ct) )

    orderedPairs = [orderBreakpoints(bp1,bp2) for bp1,bp2 in bkptPairs]
    return orderedPairs

def addBkptToDictDict(bkpt, ld):
    if not bkpt is None:
        ld[bkpt] = 1

def bkptInDictDict(bkpt, ld):
    if not bkpt is None:
        return bkpt in ld
    else:
        return False

def vcftobkpts(infile, outfile, width):
    firstbkpts = loc.locationdict(width)
    pairbkpts  = loc.locationdict(width)

    reader = pysam.VariantFile(infile)
    for record in reader.fetch():
        if list(record.filter) is None or "PASS" in list(record.filter) or list(record.filter)[0] == "." or len(list(record.filter)) == 0:
            bkptPairs = breakpointsFromRecord(record)
            for pair in bkptPairs:
                addBkptToDictDict(pair[0], firstbkpts)
                addBkptToDictDict(pair[1], pairbkpts)

    # count how many breakpoints weren't in both dicts, for diagnostics;
    # then add everything to first dict for outputting
    nunmatched = 0
    for bp in firstbkpts:
        if not bkptInDictDict(bp, pairbkpts):
            nunmatched += 1

    for bp in pairbkpts:
        if not bkptInDictDict(bp, firstbkpts):
            nunmatched += 1
            addBkptToDictDict(bp, firstbkpts)

    print("#Num breakpoints not in both lists:",nunmatched,file=sys.stderr)

    # now output everything
    for bp in firstbkpts:
        chrom, pos, strand, extendsRight = bp.asTuple()
        start = pos-width/2
        if start < 0:
            start = 0
        print("{0}	{1}	{2}".format(chrom, int(start), int(pos)+int(width/2)), file=outfile)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('infile', nargs='?', default=sys.stdin)
    parser.add_argument('outfile', nargs='?', type=argparse.FileType('w'), default=sys.stdout)
    defwidth=300
    parser.add_argument('-w','--width', type=int, help="width of breakpoint region: default("+str(defwidth)+")",default=defwidth)

    args = parser.parse_args()
    sys.exit( vcftobkpts(args.infile,args.outfile,args.width ) )



# based on https://github.com/pypa/sampleproject
from codecs import open
from distutils.command.sdist import sdist as _sdist
from distutils.core import setup
from distutils.extension import Extension
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

try:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
else:
    use_cython = True

cmdclass = {}

if use_cython:
    ext_modules = cythonize("mergesvvcf/*pyx")
    cmdclass.update({"build_ext": build_ext})
else:
    ext_modules = [
        Extension("mergesvvcf.locations", ["mergesvvcf/locations.c"]),
        Extension("mergesvvcf.variantdict", ["mergesvvcf/variantdict.c"]),
        Extension("mergesvvcf.vcftobreakpoints", ["mergesvvcf/vcftobreakpoints.c"]),
    ]


class sdist(_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize

        cythonize(["mergesvvcf/*.pyx"])
        _sdist.run(self)


cmdclass["sdist"] = sdist

# Get the long description from the relevant file
with open(path.join(here, "DESCRIPTION.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="mergesvvcf",
    version="1.0.3",
    description="Merge SV VCF calls. Fork of https://github.com/ljdursi/mergevcf by Jonathan Dursi (Jonathan.Dursi@oicr.on.ca)",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/papaemmelab/mergeSVvcf",
    # Author details
    author="Max Levine",
    author_email="levinem1@mskcc.org",
    # Choose your license
    license="GPL",
    classifiers=[
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    keywords="merge sv vcfs",
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    install_requires=["pysam>=0.15.2", "Cython>=0.29.7"],
    # ext_modules=[],
    test_suite="tests",
    extras_require={"dev": ["check-manifest"], "test": ["coverage"]},
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #    package_data={
    #        'sample': ['package_data.dat'],
    #    },
    entry_points={"console_scripts": ["mergesvvcf=mergesvvcf:main"]},
    ext_modules=ext_modules,
    cmdclass={"sdist": sdist},
)



from __future__ import print_function
import pyximport

pyximport.install()
import os
import operator
import pysam
import mergesvvcf.variantdict as variantdict


def mapped_to_chromosome(chrom):
    """
    Returns true if mapped to, eg, chr1 or X;
    false if mapped to other contig, eg GL*, MT*, hs*, M*
    """
    if chrom[0:2] in ["GL", "MT", "hs", "NC"] or chrom[0:1] == "M":
        return False
    return True


def int_if_possible(val):
    """
    Returns integer value of s if conversion succeeds, else s.
    """
    if type(val) is list or type(val) is tuple:
        val = val[0]
    try:
        i = int(val)
    except ValueError:
        i = val
    return i

    # def make_info_dict(callers, records, pos1, pos2):
    #     """ generate 'median' results for info fields from all records """
    #     callermap = {
    #         "delly": {"fields": ["DV", "RV"], "tumor": 0},
    #         "svaba": {"fields": ["SR", "DR"], "tumor": 1},
    #         "gridss": {"fields": ["RP", "SR"], "tumor": 1},
    #         "brass": {"fields": ["PS", "RC"], "tumor": 1},
    #         "smoove": {"fields": ["SR", "PE"], "tumor": 0},
    #     }
    #     # fields = ['CHR2', 'END', 'SVTYPE', 'SVLEN'] + ["SR", "DR"] + ["DV", "RV"] + ["RP"]
    #     fields = [x["fields"] for x in callermap.values()]
    #     fields = [item for sublist in fields for item in sublist]
    #     info = {}
    #     for field in fields:
    #         answers = []
    #         for caller, record in zip(callers, records):
    #             if caller in callermap.keys() and field in callermap[caller]["fields"]:
    #                 if field in record.format:
    #                     answers.append([caller,int_if_possible(record.samples[callermap[caller]["tumor"]][field])])
    #                 elif field in record.info:
    #                     answers.append([caller,int_if_possible(record.info[field])])
    #                 # elif field in record.format:
    #                 #     answers.append([caller,int_if_possible(record.samples[callermap[caller]["tumor"]][field])])
    #             nanswers = len(answers)
    #         if nanswers > 0:
    #             # sorted_answers = sorted(answers)
    #             # # doesn't quite deal with even #s correctly - can't average strings
    #             # median_pos = int(nanswers / 2)
    #             # median_answer = sorted_answers[median_pos]
    #             # if not median_answer == 0:
    #             #     info[field] = median_answer
    #             for a in answers:
    #                 info[a[0] + "_" + field] = a[1]

    if "SVTYPE" in info and info["SVTYPE"] in ["DUP", "DUP:TANDEM", "DEL", "INV"]:
        if not "SVLEN" in info:
            info["SVLEN"] = pos2 - pos1
    return info


def bkptRefAltFromPair(loc1, loc2, refstr="N"):
    alt_after = loc1.__right__ == False

    if loc2 is None or loc2.__chrom__ is None:
        bkptstr = "."
    else:
        if loc2.__right__:
            delim = "["
        else:
            delim = "]"

        assert loc2.__strand__ == (loc1.__right__ != loc2.__right__)
        bkptstr = "%s%s:%d%s" % (delim, loc2.__chrom__, loc2.__pos__, delim)

    if alt_after:
        altstr = "%s%s" % (refstr, bkptstr)
    else:
        altstr = "%s%s" % (bkptstr, refstr)

    return refstr, altstr


def getSVTYPE(chr1, chr2, extend1, extend2):
    """Get SVTYPE from extend right"""
    if chr1 != chr2:
        return "TRA"

    eventmap = {True: {True: "INV", False: "DUP"}, False: {True: "DEL", False: "INV"}}

    return eventmap[extend1][extend2]


def merge(
    filenames,
    programs,
    forceSV,
    outfile,
    slop=0,
    verbose=True,
    output_ncallers=False,
    min_num_callers=0,
    filterByChromosome=True,
    noFilter=False,
    debug=False,
):
    """Merge several VCFs from different programs into a new VCF file."""

    outvcf = pysam.VariantFile(os.devnull, "w")
    if min_num_callers > 0:
        outvcf.header.filters.add(
            "LOWSUPPORT", None, None, "Not called by enough callers in ensemble"
        )

    # Returns true if the variant is PASS in the VCF file
    def passed_variant(record):
        """Did this variant pass?"""
        return (
            "PASS" in list(record.filter) or len(list(record.filter)) == 0 or noFilter
        )

    # def infoString(callers, infodict):
    #     """
    #     Generate an INFO string from the INFO dictionary plus
    #     the list of callers.
    #     """
    #     info = {}
    #     # fields = ['SVTYPE', 'SVLEN']
    #     # fields = ["svaba_SR", "svaba_DR"] + ["delly_DV", "delly_RV"] + ["gridss_RP", "gridss_SR"]
    #     # for field in fields:
    #     for field in infodict.keys():
    #         if field in infodict:
    #             res = infodict[field]
    #             if type(res) is list:
    #                 res = res[0]
    #             info[field] = res
    #     if output_ncallers:
    #         info["NumCallers"] = int(len(set(callers)))
    #     info["Callers"] = ",".join(list(set(callers)))
    #     return info

    samplemap = {}  # the samples in each file
    infomap = {}  # the info fields in each file
    formatmap = {}  # the format fields in each file
    outsamples = set()  # standardized samples

    calldict = variantdict.variantmap(awindow=0, svwindow=slop)
    for (infile, program) in zip(filenames, programs):
        count = 0
        try:
            vcf_reader = pysam.VariantFile(infile)

            # Add contigs from all files
            for contig in sorted(vcf_reader.header.contigs):
                try:
                    outvcf.header.contigs.add(contig)
                except ValueError:
                    pass

            outvcf.header.add_meta(
                key="Caller", items=[("ID", program), ("File", infile)]
            )

            # Add filters from all files
            for hfilter in vcf_reader.header.filters.itervalues():
                if hfilter.name not in ["PASS", "LOWSUPPORT"]:
                    outvcf.header.filters.add(
                        "{}_{}".format(program, hfilter.name),
                        None,
                        None,
                        hfilter.description,
                    )
            # Add info fields from all files
            infos = []
            for info in vcf_reader.header.info.itervalues():
                if info.name not in ["SVTYPE", "SVLEN", "END", "CHR2", "NumCallers"]:
                    outvcf.header.info.add(
                        "{}_{}".format(program, info.name),
                        info.number,
                        info.type,
                        info.description,
                    )
                    infos.append(info.name)
            infomap[program] = infos
            # Add format fields from all files
            formats = []
            for hformat in vcf_reader.header.formats.itervalues():
                num = None
                if hformat.name == "GT" or hformat.number == "G":
                    num = "G"
                    continue  # Genotype is special cant do multiple
                outvcf.header.formats.add(
                    "{}_{}".format(program, hformat.name),
                    num or hformat.number,
                    hformat.type,
                    hformat.description,
                )
                formats.append(hformat.name)
            formatmap[program] = formats
            # Collect samples all files
            samples = {}
            for meta in vcf_reader.header.records:
                if meta.key == "SAMPLE" and "SampleName" in meta.keys():
                    samples[meta["ID"]] = meta["SampleName"]
            if not samples:
                for samp in vcf_reader.header.samples:
                    samples[samp] = samp.rsplit("/")[-1].replace(".bam", "")
            samplemap[program] = samples
            outsamples.update(samples.values())

            for record in vcf_reader.fetch():

                if not passed_variant(record):
                    continue

                if filterByChromosome and not mapped_to_chromosome(record.chrom):
                    continue

                if verbose:
                    if count == 0:
                        print(record, program)
                    count += 1
                    if count == 100:
                        count = 0

                calldict.addrecord(record, program, forceSV)
        except (RuntimeError, TypeError, NameError, AttributeError) as e:
            if debug:
                raise (e)
            else:
                pass

    # Add samples
    for samp in sorted(outsamples):
        outvcf.header.samples.add(samp)

    # Write the results in a master vcf file for the sample
    outvcf.header.info.add("Callers", ".", "String", "Callers that made this call")

    if output_ncallers:
        outvcf.header.info.add(
            "NumCallers", 1, "Integer", "Number of callers that made this call"
        )

    outvcf.header.info.add(
        "CHR2",
        "1",
        "String",
        "Chromosome for END coordinate in case of a translocation",
    )
    outvcf.header.info.add(
        "END", "1", "Integer", "End position of the structural variant"
    )
    outvcf.header.info.add("SVTYPE", "1", "String", "Type of structural variant")
    outvcf.header.info.add("SVLEN", "1", "String", "Length of structural variant")
    outvcf.header.info.add(
        "STRANDS",
        "1",
        "String",
        "Strand orientation of the adjacency in BEDPE format (DEL:+-, DUP:-+, INV:++/--)",
    )

    outfile = open(outfile,'w')
    print(outvcf.header, end="", file=outfile)

    for variant in sorted(calldict, key=operator.itemgetter(0)):
        callers = variant[2]
        num_callers = len(set(callers))
        passes = num_callers >= min_num_callers
        filterstring = "." if passes else "LOWSUPPORT"

        if len(variant) == 3:  # snv/indel
            loc, allele, callers = variant
            if allele is None:
                print("Allele is none: loc, allele, callers = ", loc, allele, callers)
                continue
            chrom, pos, _, _ = loc.asTuple()
            vcfrec = outvcf.new_record()
            vcfrec.contig = chrom
            vcfrec.pos = pos
            vcfrec.ref = allele[0]
            vcfrec.alts = [allele[1]]
            vcfrec.filter.add(filterstring)
            vcfrec.info["Callers"] = ",".join(list(set(callers)))
            if output_ncallers:
                vcfrec.info["NumCallers"] = str(len(set(callers)))
            print(vcfrec, end="", file=outfile)
        else:
            loc1, loc2, callers, medianPos1, medianPos2, recordscalled = variant
            if filterByChromosome and not mapped_to_chromosome(loc2.chrom):
                continue

            # records = [r for c, r in recordscalled]

            avgloc1 = loc1.withPos(medianPos1)
            avgloc2 = loc2.withPos(medianPos2)
            extend1, extend2 = avgloc1.__right__, avgloc2.__right__
            ref, alt = bkptRefAltFromPair(avgloc1, avgloc2)
            vcfrec = outvcf.new_record()
            try:
                vcfrec.contig = avgloc1.chrom
                vcfrec.pos = avgloc1.pos
                vcfrec.ref = ref
            except:
                continue
            vcfrec.alts = [alt]
            vcfrec.filter.add(filterstring)
            # for key, val in sorted(infoString(callers, make_info_dict(callers, records, medianPos1, medianPos2)).items()):
            #     vcfrec.info.__setitem__(key, val)
            vcfrec.info["Callers"] = ",".join(list(set(callers)))
            if output_ncallers:
                vcfrec.info["NumCallers"] = len(set(callers))
            vcfrec.info["CHR2"] = avgloc2.chrom
            vcfrec.stop = avgloc2.pos
            svtype = getSVTYPE(avgloc1.chrom, avgloc2.chrom, extend1, extend2)
            vcfrec.info["SVTYPE"] = svtype
            if avgloc1.chrom == avgloc2.chrom:
                vcfrec.info["SVLEN"] = str(avgloc2.pos - avgloc1.pos)
            strandmap = {False: "+", True: "-"}
            vcfrec.info["STRANDS"] = strandmap[extend1] + strandmap[extend2]
            for caller, rec in recordscalled:
                if noFilter:
                    for fil in rec.filter:
                        if not fil in [".", "PASS"]:
                            vcfrec.filter.add("{}_{}".format(caller, fil))
                for info in infomap[caller]:
                    try:
                        vcfrec.info["{}_{}".format(caller, info)] = rec.info[info]
                    except KeyError:
                        continue
                    except:
                        print(info, rec.info[info])
                        raise
                for samp in samplemap[caller].items():
                    for form in formatmap[caller]:
                        try:
                            vcfrec.samples[samp[1]][
                                "{}_{}".format(caller, form)
                            ] = rec.samples[samp[0]][form]
                        except KeyError:
                            continue
                        except:
                            print(form, rec.samples[samp[0]][form])
                            raise
            print(vcfrec, end="", file=outfile)
            if debug:
                for caller, rec in recordscalled:
                    print("#" + str(rec).rstrip() + " (" + caller + ")", file=outfile)

    outvcf.close()


def readMergedCalls(infile, filterByChromosome=True, readINFO=False, skipcallers=None):
    """Read a merged callset, and return:
        - dictionary: caller name -> caller idx
        - callsets(list of lists): [calleridx][callidx]
        - calls: callidx -> record from merged"""
    invcf = pysam.VariantFile(infile)
    callerIdx = 0
    callIdx = 0
    callsets = []
    callIdxToCall = []
    callerIdxDict = {}

    if skipcallers is None:
        skipcallers = []

    for rec in invcf.fetch():
        ncalledthis = 0
        if filterByChromosome and not mapped_to_chromosome(rec.chrom):
            continue
        callers = [c for c in rec.info["Callers"] if not c in skipcallers]
        called = []
        for caller in callers:
            if not (caller in called) and not (caller in skipcallers):
                called.append(caller)

                if not caller in callerIdxDict:
                    callerIdxDict[caller] = callerIdx
                    callerIdx += 1
                    callsets.append([])

                callsets[callerIdxDict[caller]].append(callIdx)
                ncalledthis += 1

        assert len(called) == ncalledthis
        if ncalledthis > 0:
            chrom = rec.chrom
            posstart = rec.pos
            callIdxToCall.append(
                (
                    len(called),
                    chrom,
                    posstart,
                    str(rec.REF),
                    str(list(rec.alts)[0]),
                    ",".join(called),
                )
            )
            callIdx += 1

    return callerIdxDict, callsets, callIdxToCall
